# -*- coding: utf-8 -*-
"""Parsing CLI arguments.
"""
import os
import argparse

from hac.commands import app_commands


"""Parse arguments used in CLI parser.
"""
_pargs_pack_cli = [
    {
        "names": ("location",),
        "params": {
            "help":
"""Contest or problem identifier. It can be either:
  - contest/problem URL
  - string of form "website-ID/contest-ID"
""",
            "metavar": "(CONTEST | PROBLEM)",
        }
    },
    {
        "names": ("problems",),
        "params": {
            "action": "append",
            "nargs": "*",
            "help":
"""Optional identifiers for additional problems from
contest specified by previous argument. Identifiers
can be either:
  - problem's ID
  - problem's index (counting from 1)
""",
            "metavar": "PROBLEM"
        }
    },
]


def get_pargs_pack_cli():
    """Returns CLI parse arguments.
    """
    return _pargs_pack_cli


# Parser notes.
_parser_cli_description = \
"""To execute special command that doesn't fetch remote data run: 

    hac (--help | --version | --copy-config)

To execute command that fetches remote data and processes it run:

    hac [options...] ({0}) (CONTEST | PROBLEM) [PROBLEM [PROBLEM ...]]

""".format(" | ".join(sorted(app_commands.keys())))

_parser_cli_epilog = \
"""examples: https://github.com/plesiv/hac#examples
 
"""


def get_bare_cli_parser():
    """Returns bare CLI parser object without arguments. Arguments should be
    added to parser manually.
    """
    return argparse.ArgumentParser(
        description = _parser_cli_description,
        epilog = _parser_cli_epilog,
        formatter_class=argparse.RawTextHelpFormatter
    )




# -*- coding: utf-8 -*-
"""hac - Helper for Algorithm Competitions
"""
import os
import sys


__author__ = 'Zoran Plesivčak'
__version__ = '1.0.1'
__license__ = 'GPLv2'


# -- Enums --------------------------------------------------------------------
class DataType(object):
    """Data types."""
    LANG = 0
    RUNNER = 1
    SITE = 2


class ExitStatus(object):
    """Exit status code constants."""
    OK = 0
    ERROR = 1


# -- Settings -----------------------------------------------------------------

# Application constant settings.
SETTINGS_CONST = {
    "hac_root_path": os.path.abspath(os.path.dirname(__file__)),
    "config_filename": "hacrc",
    "config_dir": "config",
    "config_user_path": os.environ.get('HAC_CONFIG_DIR',
                                       os.path.expanduser('~/.config/hac')),
    "plugin_dir": {
        DataType.LANG: "lang",
        DataType.RUNNER: "runner",
        DataType.SITE: "site",
    },
    "plugin_temp": {
        DataType.LANG: "temp",
        DataType.RUNNER: "temp",
    },
    "plugin_temp_sep": {
        DataType.LANG: ".",
        DataType.RUNNER: ".",
    },
    "plugin_temp_part_prefix": {
        DataType.RUNNER: r"\$",
    },
}




# -*- coding: utf-8 -*-
"""Common functionality for parsing CLI/configuration.
"""
import os

import hac
from hac import DataType
from hac.commands import app_commands, app_commands_help


_commands_choices = sorted(app_commands.keys())
_commands_help = "\n".join(["Command to be executed:"] +
                            [app_commands_help[com]
                                for com in sorted(app_commands.keys())])


"""Constant common parse arguments used in:

    - parsers for CLI files,
    - parsers for configuration files.
"""
_pargs_pack_common_const = [
    {
        "names": ("--version",),
        "params": {
            "action": "store_true",
            "help": """show application version and exit""",
            "dest": "version",
            "default": False,
        },
    },
    {
        "names": ("--copy-config",),
        "params": {
            "action": "store_true",
            "help":
"""copy configuration to user's local directory

""",
            "dest": "copy_config",
            "default": False,
        },
    },
    (
        {
            "names": ("-v", "--verbose"),
            "params": {
                "action": "store_true",
                "help": """increase output verbosity""",
                "dest": "verbose",
                "default": False,
            },
        },
        {
            "names": ("-V", "--terse"),
            "params": {
                "action": "store_true",
                "help": """decrease output verbosity""",
                "dest": "quiet",
                "default": False,
            },
        },
    ),
    {
        "names": ("-d", "--subdir-depth"),
        "params": {
            "type": int,
            "choices": [0, 1, 2],
            "help":
"""Additional levels at which to create files:
  0 - working directory,
  1 - directory for each problem,
  2 - directory for contest and sub-directory every
      task""",
            "dest": "subdir_depth"
        }
    },
    {
        "names": ("-w", "--workdir"),
        "params": {
            "help":
"""Change working directory (it has to exist).
Current directory by default.""",
            "metavar": "WD",
            "dest": "workdir"
        }
    },
    {
        "names": ("-t", "--tests"),
        "params": {
            "type": int,
            "choices": [0, 1],
            "help":
"""Prepare testcases (I/O samples):
  0 - no testcases
  1 - pretests""",
            "dest": "tests"
        }
    },
    (
        {
            "names": ("-f", "--force"),
            "params": {
                "action": "store_true",
                "help": "Overwrite existing files",
                "dest": "force"
            }
        },
        {
            "names": ("-F", "--no-force"),
            "params": {
                "action": "store_false",
                "help": "Warn if files already exist",
                "dest": "force"
            }
        },
    ),
    {
        "names": ("command",),
        "params": {
            "nargs": "?",
            "choices": _commands_choices,
            "help": _commands_help,
            "metavar": " | ".join(_commands_choices)
        }
    },
]

def get_pargs_pack_common(choice_langs = [], choice_runners = []):
    """Returns merge of constant and variable common parse arguments.
    """
    _pargs_pack_common_var = [
        {
            "names": ("-l", "--lang"),
            "params": {
                "action": "append",
                "choices": ['no'] + choice_langs,
                "help":
                    """Language-templates to copy""",
                "dest": "lang"
            }
        },
        {
            "names": ("-r", "--runner"),
            "params": {
                "action": "append",
                "choices": ['no'] + choice_runners,
                "help": """Runner-templates to process and copy""",
                "dest": "runner"
            }
        },
    ]

    return _pargs_pack_common_const + _pargs_pack_common_var


def pargs_packed_add(parser, pack):
    """Unpacks parse arguments and adds them to parser.

    Each element in pack is either:

      - dict to be unpacked directly in `add_argument`
      - tuple containing dicts of arguments that should be put in
        mutually-exclusive group

    """
    for arg in pack:
        # Unpack mutually-exclusive group (tuple).
        if type(arg) == tuple:
            group = parser.add_mutually_exclusive_group()
            for subarg in arg:
                group.add_argument(*subarg["names"], **subarg["params"])

        # Unpack regular argument.
        else:
            parser.add_argument(*arg["names"], **arg["params"])




# -*- coding: utf-8 -*-
"""This module provides the main functionality.
"""
import os
import sys
import textwrap
from os.path import dirname, realpath

import hac
from hac import DataType, ExitStatus
from hac.commands import app_commands
from hac.parse_common import get_pargs_pack_common, pargs_packed_add
from hac.parse_config import get_bare_config_parser
from hac.parse_cli import get_pargs_pack_cli, get_bare_cli_parser
from hac.util_common import error, dict_override, list_reduce, mainargs_index,\
    choice_generate, choice_normal, safe_cpdir
from hac.util_data import plugin_collect, plugin_match_site


def main(args=sys.argv[1:]):
    """Execution flow of the application:

        1) discover and load plugins (languages, runners and sites)
            1a] read application global defaults
            1b] override with user's custom plugins
        2) handle configuration (hacrc files and CLI arguments)
            2a] read hacrc from application global defaults
            2b] override with user's hacrc
            2c] override with command-line arguments
        3) handle special command-line switches
        4) fetch data from site
        5) execute command (prep, show)
    """

    # -- PLUGIN-SYSTEM -------------------------------------------------------
    config_global_path = os.path.join(hac.SETTINGS_CONST["hac_root_path"],
                                      hac.SETTINGS_CONST["config_dir"])
    config_user_path = hac.SETTINGS_CONST["config_user_path"]
    # User configs override global configs
    config_paths = [config_user_path, config_global_path]

    # Discover plug-ins (sites) and templates (runners, languages).
    plugin_langs = plugin_collect(config_paths, DataType.LANG)
    plugin_runners = plugin_collect(config_paths, DataType.RUNNER)
    plugin_sites = plugin_collect(config_paths, DataType.SITE)

    # Auxiliary data. If
    #   available_langs == ['cpp.0', 'cpp.1', 'py.15']), then
    #   choice_langs    == ['cpp', 'cpp.0', 'cpp.1', 'py', 'py.15']
    available_langs = plugin_langs.keys()
    sep_langs = hac.SETTINGS_CONST['plugin_temp_sep'][DataType.LANG]
    choice_langs = choice_generate(available_langs, sep_langs)

    available_runners = plugin_runners.keys()
    sep_runners = hac.SETTINGS_CONST['plugin_temp_sep'][DataType.RUNNER]
    choice_runners = choice_generate(available_runners, sep_runners)


    # -- READ CONFIG FILES ---------------------------------------------------
    # Construct and use parsers (utilize data from plug-ins).

    # Get parser arguments.
    pargs_pack_common = get_pargs_pack_common(
            choice_langs = choice_langs,
            choice_runners = choice_runners,
    )
    pargs_pack_cli = get_pargs_pack_cli()

    # Add arguments to parsers.
    parser_config = get_bare_config_parser()
    parser_cli = get_bare_cli_parser()

    pargs_packed_add(parser_config, pargs_pack_common)
    pargs_packed_add(parser_cli, pargs_pack_common)
    pargs_packed_add(parser_cli, pargs_pack_cli)

    # Get default application configuration (from files).
    config_global_file = os.path.join(hac.SETTINGS_CONST["hac_root_path"],
                                      hac.SETTINGS_CONST["config_dir"],
                                      hac.SETTINGS_CONST["config_filename"])
    assert os.path.exists(config_global_file)
    env_global = parser_config.parse_args(['@' + config_global_file])
    conf_global = vars(env_global)

    # Get user specific configuration (from files).
    config_user_file = os.path.join(hac.SETTINGS_CONST["config_user_path"],
                                    hac.SETTINGS_CONST["config_filename"])

    if os.path.exists(config_user_file):
        env_user = parser_config.parse_args(['@' + config_user_file])
        # Resolve configuration read from files
        conf_user = dict_override(conf_global, vars(env_user))
    else:
        conf_user = conf_global


    # -- READ CLI ------------------------------------------------------------
    # Special handling of CLI arguments.

    # Print help message and exit if:
    #
    #   - no arguments given OR
    #   - "-h" or "--help" is given as optional argument
    #
    if (len(args) == 0) or any([o in args for o in ("-h", "--help")]):
         parser_cli.print_help()
         sys.exit(ExitStatus.OK)

    # Print application version and exit if:
    #
    #   - "--version" is given as optional argument
    #
    if "--version" in args:
        print(
        "hac v{0}, License {1}, Copyright (C) 2014-2015 {2}".format(
                                    hac.__version__,
                                    hac.__license__,
                                    hac.__author__))
        sys.exit(ExitStatus.OK)

    # Copy global configuration to user's local directory if:
    #
    #   - "--copy-config" is given as optional argument
    #
    if "--copy-config" in args:
        safe_cpdir(os.path.join(hac.SETTINGS_CONST["hac_root_path"],
                                hac.SETTINGS_CONST["config_dir"]),
                   hac.SETTINGS_CONST["config_user_path"])

        sys.exit(ExitStatus.OK)

    # Use default command (from configuration files) if:
    #
    #   - no command given
    #
    margs_ind = mainargs_index(args)
    if (margs_ind == len(args)) or (args[margs_ind] not in app_commands):
        args.insert(margs_ind, conf_user["command"])

    # Notify user and exit if:
    #
    #   - no location (legal CONTEST / PROBLEM) given
    #
    if (len(args) == margs_ind + 1) and (args[margs_ind] in app_commands):
        error("No CONTEST / PROBLEM given!")
        sys.exit(ExitStatus.ERROR)


    # Regular handling of CLI arguments. Parse CLI arguments and resolve with
    # respect to configuration files.
    env_cli = parser_cli.parse_args(args=args)
    if env_cli:
        conf_all = dict_override(conf_user, vars(env_cli))
    else:
        conf_all = conf_user

    # -- PROCESS CONFIG / CLI ------------------------------------------------
    # Normalize aggregated configuration.

    # Reduce lang, runner lists and extract problems.
    conf_all["lang"] = list_reduce(conf_all["lang"])
    conf_all["runner"] = list_reduce(conf_all["runner"])
    conf_all["problems"] = conf_all["problems"][0]

    # Normalize lang and runner lists.
    # If available_langs == ['cpp.0', 'cpp.1', 'py.15'], then
    # conf_all["runner"] == ['cpp.0', 'cpp.1', 'py'] is normalized to
    # ['cpp.0', 'py.15']
    conf_all["lang"] = choice_normal(conf_all["lang"], available_langs)
    conf_all["runner"] = choice_normal(conf_all["runner"], available_runners)

    # Normalize location member (should be URL).
    if not conf_all['location'].startswith('http://') and \
       not conf_all['location'].startswith('https://'):
        conf_all['location'] = 'http://' + conf_all['location']


    # -- FETCH / PROCESS / PREPARE DATA --------------------------------------
    # NOTE: Matching done in two steps for testability.

    # Get site processor:
    #     1) Match site-processor, gets url of matched processor.
    site_url = plugin_match_site(plugin_sites, conf_all)
    #     2) Extract site-processor.
    site_matched = [site for site in plugin_sites if site_url == site.url]
    assert site_matched
    site_obj = site_matched[0]

    # Get contest data (utilize web-site processor).
    contest_url = site_obj.match_contest(conf_all)
    contest_obj = site_obj.get_contest(contest_url)

    # Get problems data (utilize web-site processor).
    problems_urls = site_obj.match_problems(conf_all)
    problems_objs = site_obj.get_problems(problems_urls)


    # -- EXECUTE -------------------------------------------------------------
    assert conf_all["command"] in app_commands

    # Execute selected command with all relevant information
    return app_commands[conf_all["command"]](
        conf_global = conf_global,
        conf_user = conf_user,
        conf_all = conf_all,
        plugin_langs = plugin_langs,
        plugin_runners = plugin_runners,
        plugin_sites = plugin_sites,
        site_obj = site_obj,
        contest_obj = contest_obj,
        problems_objs = problems_objs)




# -*- coding: utf-8 -*-
"""Common utilities.
"""
import re
import sys
import os
import stat
import shutil
from os.path import exists, isdir
from shutil import rmtree


# -- Printing to CLI ----------------------------------------------------------
def warn(msg):
    sys.stderr.write("WARNING: " + msg + os.linesep)

def error(msg):
    sys.stderr.write("ERROR: " + msg + os.linesep)


# -- Lists and dictionaries ---------------------------------------------------
def dict_override(a, b):
    """Overriding dictionary values.

    Values in second dictionary override the values in the first one (except
    when relevant value in second dictionary is None).

    >>> dict_override({'a': 1,'b': 2}, {'a': 3,'b': None}) == {'a': 3,'b': 2}
    True

    >>> dict_override({'a': 3,'b': None}, {'a': 1,'b': 2}) == {'a': 1,'b': 2}
    True
    """
    res = {}
    for key in (set(a.keys()) | set(b.keys())):
        if (key in b) and (b[key] != None):
            res[key] = b[key]
        else:
            res[key] = a[key]
    return res


def list_reduce(a):
    """Reduces provided list in two steps:
        1) removes all elements from list that appear prior to any 'no' element,
        2) removes duplicates and sorts elements.

    >>> list_reduce(['cpp', 'no', 'php', 'py', 'php', 'py'])
    ['php', 'py']

    >>> list_reduce(['cpp', 'no', 'php', 'py', 'php', 'no', 'py', 'cpp'])
    ['cpp', 'py']
    """
    b = list(reversed(a))
    if "no" in b:
        ind = b.index("no")
        del b[ind:]
    return sorted(set(b))


# -- Command line arguments ---------------------------------------------------
def mainargs_index(a):
    """Removes optional arguments that precede mandatory arguments.

    >>> mainargs_index(['-lcpp', '--no', '-t', 'php', 'py'])
    3

    >>> mainargs_index(['--cpp', '--no'])
    2

    >>> mainargs_index([])
    0
    """
    indices = [ i for i, v in enumerate(a) if not v.startswith("-") ]
    return indices[0] if indices else len(a)


def choice_generate(a, separator='.'):
    """Returns input list with additional entries that represent choices
    without priority specifiers.

    >>> choice_generate(['no', 'cpp.0', 'cpp.1'])
    ['cpp', 'cpp.0', 'cpp.1', 'no']

    >>> choice_generate(['no', 'cpp.0', 'cpp.1', 'py.15'])
    ['cpp', 'cpp.0', 'cpp.1', 'no', 'py', 'py.15']
    """
    separator = '.'
    b = set(a)
    for e in a:
        if separator in e:
            b.add(e.split(separator)[0])
    return sorted(b)


def choice_normal(a, b, separator='.'):
    """Normalizes list a so that among multiple choices that differ only in
    priority modifier, just the highest priority one is present in the output
    list. All the members in the output list are in the canonic form
    <TYPE>.<PRIORITY>.

    If in there is entry without the priority modifier in the input list, it
    corresponds to the request for highest priority choice available (one with
    lowest priority modifier).

    >>> choice_normal(['cpp.0', 'cpp.1', 'py'], ['cpp.0', 'cpp.1', 'py.15'])
    ['cpp.0', 'py.15']

    >>> choice_normal(['cpp', 'py.1', 'py'], ['cpp.1', 'py.0', 'py.1'])
    ['cpp.1', 'py.0']

    DEFINITIONS:
        - regular entries: 'cpp', 'cpp.0', 'py', 'py.15'
        - canonic entries: 'cpp.0', 'py.15'
        - bare entries: 'cpp', 'py'
    """
    separator = '.'
    assert all([separator in ec for ec in b])
    c2c = {ec: ec for ec in b}  # Map canonic to canonic.

    r2c = c2c.copy()            # Map regular to canonic.
    for ec in sorted(set(b)):
        eb = ec.split(separator)[0]
        if eb not in r2c:
            r2c[eb] = ec

    r2b = {}                    # Map regular to bare.
    for er in r2c:
        assert separator in r2c[er]
        r2b[er] = r2c[er].split(separator)[0]

    b_track = set()
    ret = []
    for er in sorted(set(a)):
        eb = r2b[er]
        ec = r2c[er]
        if eb not in b_track:
            b_track.add(eb)
            ret.append(ec)

    return ret


# -- Templating ---------------------------------------------------------------
def indent(text, ws=''):
    """Implementation of textwrap.indent (textwrap.indent not available on
    Python 2).

    >>> indent('a\\n  \\n b', '   ')
    '   a\\n  \\n    b'
    """
    lines = [ws + ln if (ln and not ln.isspace()) else ln
                     for ln in text.rstrip().split(os.linesep)]
    return os.linesep.join(lines)


def indent_distribute(text, mapps, kprefix=r'\$'):
    """Arguments are:

        * text - template-text,
        * mapps - maps template-part name to template-part contents.

    Only template parts whose names occur as the first non-whitespace
    characters on the line in the template-text are considered. Function
    processes template-text and template-parts by:

        * distributing whitespace before template-part name found in
          template-text to the lines of corresponding template-part contents
          (this matched whitespace is removed from template-text),
        * removes everything after template-part name until the end-of-line.

    >>> indent_distribute(' $pat', {'pat': 'a\\n b\\n  \\nc'})
    ('$pat', {'pat': ' a\\n  b\\n  \\n c'})
    """
    rtext = text
    rmapps = mapps
    for key in mapps:
        pattern = re.compile(r'^(?P<ws>\s*)(?P<part>' + kprefix + key + r'\b)',
            re.MULTILINE)
        token = pattern.search(text)

        if token:
            ws = token.group('ws')
            part = token.group('part')
            rtext = re.sub(pattern, part, rtext)
            rmapps[key] = indent(mapps[key], ws)

    return rtext, rmapps


# -- Filesystem ---------------------------------------------------------------
def safe_mkdir(path, force=False):
    """Carefully handles directory creation. Notifies about special
    occurrences.

    Argument force used to decide if priorly existing file named "path" should
    be replaced with directory named "path".
    """
    if not exists(path):
        os.mkdir(path)
    else:
        if isdir(path):
            warn('Directory "' + path + '" already exists!')
        else:
            # Distinguish between two cases (depending on argument force), if:
            #
            #   - path exists and
            #   - it's not a directory
            #
            if force:
                warn('Deleting file "' + path + '" and creating directory!')
                os.remove(path)
                os.mkdir(path)
            else:
                warn('"' + path + '" is not a directory!')


def safe_cpdir(path_from, path_to, force=False):
    """Carefully handles recursive directory copying. Notifies about special
    occurrences.

    Argument force used to decide if priorly existing files in destination
    directory should be replaced with new files from source.
    """
    # When force is true (branching):
    #
    #
    # - source is DIR  && destination is DIR  -> do nothing
    #
    # - source is FILE && destination is DIR  \
    # - source is DIR  && destination is FILE  -> remove destination
    # - source is FILE && destination is FILE /
    #
    if force:
        if exists(path_from) and not isdir(path_from) and isdir(path_to):
            warn('Replacing directory "' + path_to + '"!')
            rmtree(path_to)

        if exists(path_from) and exists(path_to) and not isdir(path_to):
            warn('Replacing file "' + path_to + '"!')
            os.remove(path_to)

    # When source is directory (sequence):
    #
    # 1) destination doesn't exist -> create directory
    # 2) destination is directory  -> recursively copy
    #
    if isdir(path_from):
        if exists(path_to) and not isdir(path_to):
            warn('File named "' + path_to + '" already exists!')

        if not exists(path_to):
            os.mkdir(path_to)

        if isdir(path_to):
            for fp in os.listdir(path_from):
                safe_cpdir(os.path.join(path_from, fp),
                           os.path.join(path_to, fp),
                           force)

    # When source is file:
    #
    # - destination doesn't exist -> copy file
    #
    elif exists(path_from):
        if exists(path_to):
            warn('File/directory named "' + path_to + '" already exists!')
        else:
            shutil.copyfile(path_from, path_to)


def safe_fwrite(path, contents="", force=False, executable=False):
    """Carefully handles file writing. Notifies about special occurrences.

    Argument force used to decide if priorly existing file named "path" should
    be overwritten with provided contents.
    """
    # Path exists and is directory.
    if isdir(path):
        if force:
            warn('Deleting directory "' + path + '" and creating file!')
            rmtree(path)
        else:
            warn('Directory named "' + path + '" already exists!')

    # Path exists but is not directory.
    elif exists(path):
        if force:
            warn('Deleting file "' + path + '" and creating file!')
            os.remove(path)
        else:
            warn('File named "' + path + '" already exists!')

    # Writing to file.
    if not exists(path):
        with open(path, 'w') as f:
            f.write(contents)

    # Make file executable for everyone
    if executable:
        perms = os.stat(path)
        os.chmod(path, perms.st_mode | stat.S_IEXEC |
                                       stat.S_IXGRP |
                                       stat.S_IXOTH)


# -- Metaclassing (portable, works on Python2/Python3) ------------------------
def with_metaclass(mcls):
    def decorator(cls):
        body = vars(cls).copy()
        # Clean out class body.
        body.pop('__dict__', None)
        body.pop('__weakref__', None)
        return mcls(cls.__name__, cls.__bases__, body)
    return decorator




# -*- coding: utf-8 -*-
"""Implementation of application commands.

Commands can access following entries from dictionary "args":

    - conf_global: dictionary containing default application settings,
    - conf_user: dictionary containing user settings from configuration files
                 (override default application settings),
    - conf_all: dictionary containing command-line settings (override user
                settings from configuration files),
    - plugin_langs: dictionary mapping language template designation to the
                    contents of language template file
    - plugin_runners: dictionary (of dictionaries) mapping runner templates
                      designation to the dictionary which map language template
                      designation to the contents of processed runner template
                      file
    - plugin_sites: list of all available site-processors (objects whose classes
             inherit from hac.data.ISite),
    - site_obj: selected site-processor (object whose class inherit from
                hac.data.ISite),
    - contest_obj: selected contest (instance of hac.data.Contest),
    - problems_objs: list of selected problems (instances of hac.data.Problem).
"""
import sys
import os
from os.path import expanduser, exists, isdir, join
from pprint import PrettyPrinter

import hac
from hac import DataType, ExitStatus
from hac.util_common import warn, error, safe_mkdir, safe_fwrite
from hac.data import ISite, Contest, Problem


def _command_prep(**args):
    """Prepares the environment for selected problems.

    This command is idempotent irrespective of the "--force" switch.
    """
    conf_all = args['conf_all']
    dir_working = expanduser(conf_all['workdir'])

    # 1) Working directory has to exist.
    if not isdir(dir_working):
        error('Directory "' + dir_working + '" does not exist!')
        return ExitStatus.ERROR

    # 2) Establish contest directory.
    contest_obj = args['contest_obj']
    if conf_all['subdir_depth'] == 2:
        dir_contest = join(dir_working, contest_obj.ID)
        safe_mkdir(dir_contest, force=conf_all['force'])
    else:
        dir_contest = dir_working

    # Proceed if there exists directory hierachy until this point
    if (isdir(dir_contest)):

        # 3) Establish problems directories.
        problems_objs = args['problems_objs']
        if conf_all['subdir_depth'] >= 1:
            problems_dirs = {}
            for prob in problems_objs:
                problems_dirs[prob] = join(dir_contest, prob.ID)
                safe_mkdir(problems_dirs[prob], force=conf_all['force'])
        else:
            problems_dirs = {prob: dir_contest for prob in problems_objs}

        plugin_langs = args['plugin_langs']
        selected_langs = conf_all['lang']
        sep_langs = hac.SETTINGS_CONST['plugin_temp_sep'][DataType.LANG]

        plugin_runners = args['plugin_runners']
        selected_runners = conf_all['runner']
        sep_runners = hac.SETTINGS_CONST['plugin_temp_sep'][DataType.RUNNER]

        # 4) Create language and runner templates.
        # For each problem ...
        for prob in problems_objs:
            if isdir(problems_dirs[prob]):
                problem_path = join(problems_dirs[prob], prob.ID)

                # ... create language for all selected languages.
                for lang in selected_langs:
                    assert sep_langs in lang
                    assert lang in plugin_langs

                    lang_ext = lang.split(sep_langs)[0]

                    lang_file = problem_path + os.extsep + lang_ext
                    safe_fwrite(lang_file, plugin_langs[lang],
                                force=conf_all['force'])

                # ... create runner for every combiantion of runner/language.
                for runn in selected_runners:
                    for lang in selected_langs:
                        assert sep_runners in runn
                        assert runn in plugin_runners

                        runn_ext = runn.split(sep_runners)[0]
                        lang_ext = lang.split(sep_langs)[0]

                        if lang_ext in plugin_runners[runn]:
                            runner_file = problem_path + os.extsep + \
                                          lang_ext + os.extsep + runn_ext
                            safe_fwrite(runner_file,
                                        plugin_runners[runn][lang_ext],
                                        force=conf_all['force'],
                                        executable=True)
                        else:
                            warn("Runner for [{0}/{1}] combo can't be created!"
                                  .format(runn, lang_ext))

                # 5) Dump inputs and outputs.
                if conf_all['tests'] >= 1:
                    for i, inp in enumerate(prob.inputs):
                        in_file = join(problems_dirs[prob],
                                       prob.ID + os.extsep + str(i+1) +
                                       os.extsep + 'in')
                        safe_fwrite(in_file, inp, force=conf_all['force'])

                    for i, out in enumerate(prob.outputs):
                        out_file = join(problems_dirs[prob],
                                       prob.ID + os.extsep + str(i+1) +
                                       os.extsep + 'out')
                        safe_fwrite(out_file, out, force=conf_all['force'])

    return ExitStatus.OK


def _command_show(**args):
    """Displays information about:
        - configuration (application default and user specific),
        - command line arguments.

    Expects labels for arguments to be contained in "args_text" argument.
    """
    verbose = args['conf_all']['verbose']

    # Prepare labels for printing
    #  -> TERSE (opposite of verbose)
    args_labels = {
        'conf_all':       '1c --- Total config (overide 1b) .... ',
        'site_obj':       '3a --- Selected site processor ...... ',
        'contest_obj':    '3b --- Selected contest ............. ',
        'problems_objs':  '3c --- Selected problems ............ ',
    }

    #  -> VERBOSE
    if verbose:
        args_labels.update({
            'conf_global':    '1a --- App default config ........... ',
            'conf_user':      '1b --- User files config (override 1a)',
            'plugin_langs':   '2a --- Available language templates . ',
            'plugin_runners': '2b --- Available runner templates ... ',
            'plugin_sites':   '2c --- Available site processors .... ',
        })


    # Prepare application setting-and-config values for printing
    args_printable = {
        'conf_global': args['conf_global'],
        'conf_user': args['conf_user'],
        'conf_all': args['conf_all'],

        # args_printable['plugin_langs'] = ['cpp.0', 'cpp.1', 'py.0']
        'plugin_langs': args['plugin_langs'].keys(),

        # args_printable['plugin_langs'] = {'sh.0': ['cpp', 'py'],
        #                                   'sh.1': ['cpp', 'py']},
        'plugin_runners': {r: args['plugin_runners'][r].keys()
                           for r in args['plugin_runners']},

        # Prepare for TERSE or VERBOSE
        'plugin_sites': [{k: site.__dict__[k]
                          for k in ISite.get_props(verbose)}
                         for site in args['plugin_sites']],
        'site_obj': {k: args['site_obj'].__dict__[k]
                     for k in ISite.get_props(verbose)},
        'contest_obj': {k: args['contest_obj'].__dict__[k]
                        for k in Contest.get_props(verbose)},
        'problems_objs': [{k: prob.__dict__[k]
                           for k in Problem.get_props(verbose)}
                          for prob in args['problems_objs']]
    }

    #TODO display language x runner matrix

    # Construct dictionary with data to print
    data = {args_labels[key]: args_printable[key] for key in args_labels}

    # Pretty-print data
    printer = PrettyPrinter(indent=1, width=1)
    printer.pprint(data)

    return ExitStatus.OK


# Application commands collected in dictionary
app_commands = { "prep": _command_prep,
                 "show": _command_show }

app_commands_help = {
"prep":
"""  - prep - prepare directories and files for
    specified problems""",

"show":
"""  - show - print relevant information about
    application configuration, available-plugins,
    identified contest and problems to be fetched
    (changeable verbosity)"""
}




# -*- coding: utf-8 -*-
"""Parsing configuration files.
"""
import argparse


class ParserConfig(argparse.ArgumentParser):
    """Class defining parsers for configuration files. Ignores lines starting
    with '#', and parses all other lines as if they were given on CLI.
    """
    def convert_arg_line_to_args(self, arg_line):
        if len(arg_line) > 0 and arg_line[0] == '#':
            return
        for arg in arg_line.split():
            if not arg.strip():
                continue
            yield arg


def get_bare_config_parser():
    """Returns bare parser for configuration files. Arguments should be added
    to parser manually.
    """
    return ParserConfig(fromfile_prefix_chars='@')




# -*- coding: utf-8 -*-
"""Data-structures definitons for:
    - web-sites
    - contests
    - problems
"""
from abc import ABCMeta, abstractmethod

import hac
from hac.util_common import with_metaclass


# -- Dynamic data (plugins) ---------------------------------------------------
class ISiteRegistry(ABCMeta):
    """Dynamic registry of sites (plugin architecture).
    """
    sites = []
    def __init__(cls, name, bases, attrs):
        if name != 'ISite':
            ISiteRegistry.sites.append(cls)


@with_metaclass(ISiteRegistry)
class ISite(object):
    """Site template.
    """

    def __init__(self, name=None, ID=None, url=None, time_limit_ms=2000,
                 memory_limit_kbyte=262144, source_limit_kbyte=64):
        self.name = name
        self.ID = ID
        self.url = url
        self.time_limit_ms = time_limit_ms
        self.memory_limit_kbyte = memory_limit_kbyte
        self.source_limit_kbyte = source_limit_kbyte

    @staticmethod
    def get_props(verbose=False):
        return ['ID', 'url'] if not verbose else \
               ['name', 'ID', 'url', 'time_limit_ms', 'memory_limit_kbyte',
                'source_limit_kbyte']

    @abstractmethod
    def match_contest(self, conf):
        """Returns well formated contest URL according to user input.
        """
        pass

    @abstractmethod
    def get_contest(self, url):
        """Expects well formated URL. Returns Contest object.
        """
        pass

    @abstractmethod
    def match_problems(self, conf):
        """Returns list of well formated problem URLs according to user input.
        """
        pass

    @abstractmethod
    def get_problems(self, urls):
        """Expects list of well formated URLs. Returns list of Problem objects.
        """
        pass


# -- Containers ---------------------------------------------------------------
class Contest(object):
    """Contest info container.
    """

    def __init__(self, name=None, ID=None, url=None):
        self.name = name
        self.ID = ID
        self.url = url

    @staticmethod
    def get_props(verbose=False):
        return ['ID', 'url'] if not verbose else \
               ['name', 'ID', 'url']


class Problem(object):
    """Problem info container.
    """

    def __init__(self, name=None, ID=None, url=None, time_limit_ms=2000,
                 memory_limit_kbyte=262144, source_limit_kbyte=64, inputs=None,
                 outputs=None):
        self.name = name
        self.ID = ID
        self.url = url
        self.time_limit_ms = time_limit_ms
        self.memory_limit_kbyte = memory_limit_kbyte
        self.source_limit_kbyte = source_limit_kbyte
        self.inputs = inputs or []
        self.outputs = outputs or []

    @staticmethod
    def get_props(verbose=False):
        return ['ID', 'url'] if not verbose else \
               ['name', 'ID', 'url', 'time_limit_ms', 'memory_limit_kbyte',
                'source_limit_kbyte', 'inputs', 'outputs']




# -*- coding: utf-8 -*-

import re
import sys

if sys.version_info.major == 2:
    from urlparse import urlparse
else:
    from urllib.parse import urlparse

from hac.data import ISite, Contest, Problem


class SiteLocal(ISite):
    """Local site processor.
    """

    # URL templates.
    url_temp_cont = "http://localhost/{0}"
    url_temp_s_prob = "/{0}"

    # Regex patterns.
    patt_cont = re.compile(r"/(?P<CONT>[^/]*)?(/(?P<PROB>[^/]*))?")
    patt_prob = re.compile(r"[^/]+")


    def __init__(self):
        self.url = "localhost"
        self.name = "Local"
        self.ID = "local"
        self.time_limit_ms = None
        self.memory_limit_kbyte = None
        self.source_limit_kbyte = None


    def match_contest(self, conf):
        """Extracts contest data from conf and generates canonic URL
        identifying that contest.
        """
        loc_path = urlparse(conf['location']).path or '/'
        tokens = SiteLocal.patt_cont.search(loc_path)
        ID_cont = tokens.group('CONT') or 'contest'
        return SiteLocal.url_temp_cont.format(ID_cont)


    def get_contest(self, url):
        """Creates contest object identified by the given URL.
        """
        url_path = urlparse(url).path
        assert url_path
        cont = Contest()
        cont.url = url
        tokens = SiteLocal.patt_cont.search(url_path)
        cont.ID = tokens.group('CONT')
        assert cont.ID
        cont.name = cont.ID
        return cont


    def match_problems(self, conf):
        """Extracts problems data from conf and generates list of canonic URLs
        identifying those problems.
        """
        url_cont = self.match_contest(conf)
        url_temp_prob = url_cont + SiteLocal.url_temp_s_prob

        urls = []
        # Match single problem from 'location'.
        loc_path = urlparse(conf['location']).path or '/'
        tokens = SiteLocal.patt_cont.search(loc_path)
        ID_prob = tokens.group('PROB')
        if ID_prob:
            urls.append(url_temp_prob.format(ID_prob))

        # Match potentially multiple problems from 'problems'.
        for prob in conf['problems']:
            tokens = SiteLocal.patt_prob.findall(prob)
            ID_prob = tokens and tokens[-1]
            if ID_prob:
                urls.append(url_temp_prob.format(ID_prob))

        return urls


    def get_problems(self, urls):
        """Creates problems' objects identified by the provided list of URLs.
        """
        probs = []
        for url in urls:
            prob = Problem()
            prob.url = url
            url_path = urlparse(url).path
            assert url_path
            tokens = SiteLocal.patt_cont.search(url_path)
            prob.ID = tokens.group('PROB')
            assert prob.ID
            prob.name = prob.ID
            prob.time_limit_ms = self.time_limit_ms
            prob.memory_limit_kbyte = self.memory_limit_kbyte
            prob.source_limit_kbyte = self.source_limit_kbyte
            probs.append(prob)
        return probs




# -*- coding: utf-8 -*-

import os
import re
import sys
import requests
from lxml import html

if sys.version_info.major == 2:
    from urlparse import urlparse
else:
    from urllib.parse import urlparse

from hac.data import ISite, Contest, Problem
from hac.util_common import warn
from hac.util_data import RequestsCache


class SiteCodeforces(ISite):
    """Codeforces site processor.
    """

    # URL templates.
    url_temp_cont = "http://codeforces.com/contest/{0}"
    url_temp_s_prob = "/problem/{0}"

    # Regex patterns.
    patt_cont = re.compile(r"/(contest/)?(?P<CONT>[^/]*)?(/(problem/)?(?P<PROB>[^/]*))?")
    patt_prob = re.compile(r"[^/]+")

    # Xpath selectors.
    xpath_cont_name = '//*[@id="sidebar"]//a[contains(@href, "contest")]/text()'
    xpath_probs_IDs = '//*[@id="pageContent"]//*[@class="id"]//a/text()'
    xpath_prob_name = '//*[@id="pageContent"]//*[@class="header"]//*[@class="title"]/text()'
    xpath_prob_time = '//*[@id="pageContent"]//*[@class="time-limit"]/text()'
    xpath_prob_memory = '//*[@id="pageContent"]//*[@class="memory-limit"]/text()'
    xpath_prob_ins = '//*[@id="pageContent"]//*[@class="sample-tests"]//*[@class="input"]//pre'
    xpath_prob_outs = '//*[@id="pageContent"]//*[@class="sample-tests"]//*[@class="output"]//pre'

    # Proxy for HTTP requests (handles request caching).
    proxy = RequestsCache()

    # Helper methods
    @staticmethod
    def ID_prob_encode(id_in):
        """Codeforces problems are encoded with uppercase latin letters. This
        function handles conversion from:

            - lowercase latin letters,
            - numbers.

        >>> SiteCodeforces.ID_prob_encode('A')
        'A'

        >>> SiteCodeforces.ID_prob_encode('z')
        'Z'

        >>> SiteCodeforces.ID_prob_encode('3')
        'C'

        >>> SiteCodeforces.ID_prob_encode('.') is None
        True
        """
        if isinstance(id_in, str):

            if id_in.isalpha() and len(id_in) == 1:
                return id_in.upper()

            if id_in.isdigit():
                id_out = chr(ord('A') + int(id_in) - 1)
                if id_out.isalpha():
                    return id_out

        return None


    def __init__(self):
        self.url = "http://codeforces.com/"
        self.name = "Codeforces"
        self.ID = "codeforces"
        self.time_limit_ms = None
        self.memory_limit_kbyte = None
        self.source_limit_kbyte = 64


    def match_contest(self, conf):
        """Extracts contest data from conf and generates canonic URL
        identifying that contest.
        """
        loc_path = urlparse(conf['location']).path or '/'
        tokens = SiteCodeforces.patt_cont.search(loc_path)
        ID_cont = tokens.group('CONT') or 'contest'
        return SiteCodeforces.url_temp_cont.format(ID_cont)


    def get_contest(self, url):
        """Fetches data from the contest URL and creates contest object
        identified by that URL.
        """
        url_path = urlparse(url).path
        assert url_path
        cont = Contest()
        cont.url = url
        tokens = SiteCodeforces.patt_cont.search(url_path)
        cont.ID = tokens.group('CONT')

        # Data from web:
        #   - contest name.
        page = SiteCodeforces.proxy.get(url)
        tree = html.fromstring(page.text)
        extr = tree.xpath(SiteCodeforces.xpath_cont_name)
        cont.name = (extr and str(extr[0])) or None

        return cont


    def match_problems(self, conf):
        """Extracts problems data from conf and generates list of canonic URLs
        identifying those problems.
        """
        url_cont = self.match_contest(conf)
        url_temp_prob = url_cont + SiteCodeforces.url_temp_s_prob

        # Data from web:
        #   - available problem IDs.
        page = SiteCodeforces.proxy.get(url_cont)
        tree = html.fromstring(page.text)
        extr = tree.xpath(SiteCodeforces.xpath_probs_IDs)
        IDs_available = [ str(e.strip()) for e in extr ]

        IDs = []
        # Match single problem from 'location'.
        loc_path = urlparse(conf['location']).path or '/'
        tokens = SiteCodeforces.patt_cont.search(loc_path)
        ID_raw = tokens.group('PROB')
        ID_prob = SiteCodeforces.ID_prob_encode(ID_raw)
        if ID_prob:
            IDs.append(ID_prob)

        # Match potentially multiple problems from 'problems'.
        for prob in conf['problems']:
            tokens = SiteCodeforces.patt_prob.findall(prob)
            ID_raw = tokens and tokens[-1]
            ID_prob = SiteCodeforces.ID_prob_encode(ID_raw)
            if ID_prob:
                IDs.append(ID_prob)

        # If no problems are successfully manually selected, select them all.
        if not IDs:
            IDs = IDs_available

        # Notify about selected but non-available problems.
        urls = []
        for ID in IDs:
            if ID in IDs_available:
                urls.append(url_temp_prob.format(ID))
            else:
                warn('Problem "' + ID + '" does not exist in ' + url_cont)

        return sorted(urls)


    def get_problems(self, urls):
        """Fetches data from the problems' URLs and creates problem objects
        identified by those URLs.
        """
        probs = []
        for url in urls:
            prob = Problem()
            prob.url = url
            url_path = urlparse(url).path
            assert url_path
            tokens = SiteCodeforces.patt_cont.search(url_path)
            prob.ID = tokens.group('PROB')
            assert prob.ID
            prob.source_limit_kbyte = self.source_limit_kbyte

            # Data from web (for each problem):
            page = SiteCodeforces.proxy.get(url)
            tree = html.fromstring(page.text)
            #   - problem name,
            extr = tree.xpath(SiteCodeforces.xpath_prob_name)
            prob.name = (extr and str(extr[0])) or None
            #   - problem time limit,
            extr = tree.xpath(SiteCodeforces.xpath_prob_time)
            limit = extr and float(extr[0].split()[0]) * 1e3
            prob.time_limit_ms = limit or self.time_limit_ms
            #   - problem memory limit,
            extr = tree.xpath(SiteCodeforces.xpath_prob_memory)
            limit = extr and float(extr[0].split()[0]) * 2**10
            prob.memory_limit_kbyte = limit or self.memory_limit_kbyte
            #   - test inputs,
            extr = tree.xpath(SiteCodeforces.xpath_prob_ins)
            prob.inputs = [ os.linesep.join(inp.itertext()) for inp in extr ]
            #   - test outputs.
            extr = tree.xpath(SiteCodeforces.xpath_prob_outs)
            prob.outputs = [ os.linesep.join(out.itertext()) for out in extr ]

            probs.append(prob)
        return probs




import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
from codecs import open

import hac


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = [
            '--doctest-modules', '--verbose',
            './config/site', './hac', './tests'
        ]
        self.test_suite = True

    def run_tests(self):
        import pytest
        sys.exit(pytest.main(self.test_args))


def long_description():
    with open('README.rst', encoding='utf-8') as f:
        return f.read()

entry_points = {
    'console_scripts': [
        # When installed this way, not going through "__main__.py"
        'hac = hac.core:main',
    ],
}

config_dest='share/doc/hac/config'

setup(
    name='hac',
    version=hac.__version__,
    description=hac.__doc__.strip(),
    long_description=long_description(),
    url='https://github.com/plesiv/hac/blob/master/README.rst',
    download_url='https://github.com/plesiv/hac',
    author=hac.__author__,
    author_email='z@plesiv.com',
    license='hac.__license__',

    packages=find_packages(exclude=['tests*']),
    package_data={'hac': ['config/hacrc',
                          'config/lang/*',
                          'config/runner/*',
                          'config/site/*.py']},
    entry_points=entry_points,
    extras_require={},
    install_requires=[
        'requests>=2.3.0',
        'lxml>=3.3.5'
    ],
    tests_require=['pytest'],
    cmdclass = {'test': PyTest},

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Intended Audience :: Education',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Topic :: Education',
    ],
    keywords='algorithm competitions',
)



# -*- coding: utf-8 -*-
"""Utilities for application plugin-system.

NOTE: for simplicity, language and runner templates are referred to as plug-ins
      in this file (along with site plug-ins).
"""
import imp
import os
import sys
import re
import requests
from string import Template

if sys.version_info.major == 2:
    from urlparse import urlparse
else:
    from urllib.parse import urlparse

import hac
from hac import DataType
from hac.data import ISiteRegistry
from hac.util_common import indent_distribute


_plugin_fname_regex = {
    DataType.LANG: r"^(?P<temp>[^.]+)\.(?P<prio>[^.]+)\.(?P<ext>[^.]+)$",
    DataType.RUNNER: {
        'temp': r"^(?P<temp>[^.]+)\.(?P<prio>[^.]+)\.(?P<ext>[^.]+)$",
        'part_u': r"^(?P<lang>[^.]+)\.(?P<part>[^.]+)\.{prio}\.{ext}$",
    }
}

# -- Languages ----------------------------------------------------------------
def _plugin_discover_langs(dirs):
    """Discovers all available programming language templates in a given list
    of directories.

    Programming language template file-names are expected to be in the format:

        <PLUGIN_TEMP>.<PRIORITY>.<LANGUAGE-EXTENSION>

    Where:

        * <PLUGIN_TEMP> is application constant setting (from SETTINGS_CONST).
        * <PRIORITY> is the integer higher or equal to zero. Lower <PRIORITY>
            number indicates higher priority,
        * <LANGUAGE-EXTENSION> denotes the programming language of the
        template.

    For example, highest priority Python programming language template would
    have fname:

        temp.0.py

    Returns dictionary mapping "<LANGUAGE-EXTENSION>.<PRIORITY>" to the
    contents of the given programming language template file.
    """
    fname_pat = re.compile(_plugin_fname_regex[DataType.LANG])
    sep_l = hac.SETTINGS_CONST['plugin_temp_sep'][DataType.LANG]

    langs = {}
    for cdir in dirs:
        if os.path.isdir(cdir):
            for fname in os.listdir(cdir):
                token = fname_pat.search(fname)
                if token:
                    # Filename matches specified regular expression.
                    key = token.group("ext") + sep_l + token.group("prio")
                    if (key not in langs):
                        with open(os.path.join(cdir, fname), 'r') as f:
                            contents = f.read()
                        langs[key] = contents
    return langs


# -- Runners ------------------------------------------------------------------
def _plugin_discover_runners(dirs):
    """In a given list of directories discovers all available:

        * runner templates and
        * runner templating parts.

    Generates runners' contents by applying templating parts to the templates.
    Each template must have at least one templating part for each corresponding
    programming language.

    Runner template file-names are expected to be in the format:

        <PLUGIN_TEMP>.<PRIORITY>.<RUNNER-EXTENSION>

    Where:

        * <PLUGIN_TEMP> is application constant setting (from SETTINGS_CONST).
        * <PRIORITY> is the integer higher or equal to zero. Lower <PRIORITY>
            number indicates higher priority,
        * <RUNNER-EXTENSION> denotes runner file-type of the template.

    For example, highest priority shell runner template would have filename:

        temp.0.sh

    Runner template part file-names are expected to be in the format:

        <LANGUAGE-EXTENSION>.<PART>.<PRIORITY>.<RUNNER-EXTENSION>

    Where:

        * <LANGUAGE-EXTENSION> denotes which programming language is this
            runner template part for,
        * <PART> is the name of the runner template part,
        * <PRIORITY>.<RUNNER-EXTENSION> should correspond to the end of some
            existing <PLUGIN_TEMP>.<PRIORITY>.<RUNNER-EXTENSION> template.
            Matching to the appropriate template is done according to this
            part.

    For example, parts "compile" and "execute" for Python programming language
    and temp.0.sh template would be:

        py.compile.0.sh
        py.execute.0.sh

    Returns dictionary mapping "<RUNNER-EXTENSION>.<PRIORITY>" to the
    corresponding runner dictionaries. Each runner dictionary maps from
    "LANGUAGE-EXTENSION" to the contents of the prepared for that programming
    language.
    """
    ftemp_pat = re.compile(_plugin_fname_regex[DataType.RUNNER]['temp'])
    fpart_regex_u =_plugin_fname_regex[DataType.RUNNER]['part_u']
    sep_r = hac.SETTINGS_CONST['plugin_temp_sep'][DataType.RUNNER]
    pref_r = hac.SETTINGS_CONST['plugin_temp_part_prefix'][DataType.RUNNER]

    runners = {}
    for cdir_r in dirs:
        if os.path.isdir(cdir_r):
            for fname_r in os.listdir(cdir_r):

                # Is filename in proper runner-template format?
                tok_r = ftemp_pat.search(fname_r)
                if tok_r:
                    ext = tok_r.group("ext")
                    prio = tok_r.group("prio")

                    key_r = ext + sep_r + prio
                    fpart_regex = fpart_regex_u.format(ext=ext, prio=prio)
                    fpart_pat = re.compile(fpart_regex)
                    fpath_r = os.path.join(cdir_r, fname_r)
                    with open(fpath_r, 'r') as f:
                        contents_r = f.read()

                    # Take first occurrence of runner template.
                    if key_r not in runners:

                        # Get available parts (get the first occurence).
                        parts = {}
                        for cdir_p in dirs:
                            if os.path.isdir(cdir_p):
                                for fname_p in os.listdir(cdir_p):
                                    # Is filename in proper part format?
                                    tok_p = fpart_pat.search(fname_p)
                                    if tok_p:
                                        lang = tok_p.group("lang")
                                        part = tok_p.group("part")
                                        fpath_p = os.path.join(cdir_p, fname_p)
                                        with open(fpath_p, 'r') as f:
                                            contents_p = f.read()

                                        if lang not in parts:
                                            parts[lang] = {}

                                        if part not in parts[lang]:
                                            parts[lang][part] = contents_p

                        # Do the templating.
                        langs = {}
                        for lang in parts:
                            rtemp, rparts = indent_distribute(contents_r,
                                                              parts[lang],
                                                              pref_r)
                            template = Template(rtemp)
                            rendered = template.safe_substitute(rparts)

                            if lang not in langs:
                                langs[lang] = rendered

                        runners[key_r] = langs
    return runners


# -- Sites --------------------------------------------------------------------
def _plugin_discover_sites(dirs):
    """Dynamically discovers all web-site processors in a given list of
    directories.

    Returns list of site-processor objects. When site processors in different
    locations have the same filename, site processors occurring in an location
    specified earlier in the input list take precedence.

    All discovered classes are instantiated with empty constructor.
    """
    registered = set()
    for cdir in dirs:
        if os.path.isdir(cdir):
            for filename in os.listdir(cdir):
                froot, fext = os.path.splitext(filename)
                if (fext == ".py") and (froot not in registered):
                    fname, fpath, fdescr = imp.find_module(froot, [cdir])
                    if fname:
                        # Register discovered site-processor module in
                        # ISiteRegistry.sites
                        mod = imp.load_module(froot, fname, fpath, fdescr)
                        # Track registered sites
                        registered.add(froot)
    return [ site() for site in ISiteRegistry.sites ]


# TODO crowd-source writing of this function
def plugin_match_site(sites, conf): #very-stupid matching now
    """Must return a site. Reasonable default site if can't match any site
    explicitly.
    """
    for site in sites:
        hostname = urlparse(conf['location']).hostname.lower()
        if(hostname in site.name.lower() or hostname in site.ID.lower()):
            return site.url
    return list(sorted(sites))[0].url


# -- Common data utilities ----------------------------------------------------
_plugin_discover_funcs = {
    DataType.LANG: _plugin_discover_langs,
    DataType.RUNNER: _plugin_discover_runners,
    DataType.SITE: _plugin_discover_sites,
}


def plugin_collect(paths, data_type):
    """Collects plug-ins of specified type from the list of directories.

    IMPORTANT: Paths that appear earlier in the list take precedence over the
    paths that appear later, i.e. if the plug-in appears in multiple locations,
    one that appears in the location specified earlier will be collected
    (others will be ignored).
    """
    plugin_dir = hac.SETTINGS_CONST["plugin_dir"][data_type]
    plugin_discover = _plugin_discover_funcs[data_type]

    return plugin_discover([os.path.join(path, plugin_dir) for path in paths])


# -- Web-data utilities -------------------------------------------------------
class RequestsCache(object):

    def __init__(self):
        self._store = {}

    def get(self, url):
        if url not in self._store:
            self._store[url] = requests.get(url)
        return self._store[url]

